/*
    remote-sensor, a wireless sensor data collection system
    Copyright (C) 2008  christian mock <cm@tahina.priv.at>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

/* this file is derived from obdev's remote-sensor file of the same name */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usb.h>    /* this is libusb, see http://libusb.sourceforge.net/ */

#define USBDEV_SHARED_VENDOR    0x16C0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   0x05DC  /* Obdev's free shared PID */
/* Use obdev's generic shared VID/PID pair and follow the rules outlined
 * in firmware/usbdrv/USBID-License.txt.
 */

#define PSCMD_ECHO   0
#define PSCMD_STATUS 1
#define PSCMD_DATA   2
#define PSCMD_STATS  3
/* These are the vendor specific SETUP commands implemented by our USB device */

char *statusbits[] =
  {"FFIT", "POR", "FFOV", "WKUP", "EXT", "LBD", "FFEM", "RSSI",
   "DQD", "CRL", "ATGL", "OFFS(6)" };

char *states[] =
  { "INIT", "IDLE", "READING", "READING_DONE" };

struct stats {
  uint16_t ok;			// packets received OK
  uint16_t e_cksum;		// checksum errors
  uint16_t e_to_reading;	// timeouts in READING
  uint16_t e_to_reading_done;	// timeouts in READING_DONE
  uint16_t e_bad_len;		// len out of range
};

struct data {
  uint8_t type;
  uint8_t sender;
  uint8_t seq;
  uint16_t voltage;
} __attribute((packed));


static void usage(char *name)
{
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "  %s status\n", name);
    fprintf(stderr, "  %s test\n\n", name);
}


static int  usbGetStringAscii(usb_dev_handle *dev, int index, int langid, char *buf, int buflen)
{
char    buffer[256];
int     rval, i;

    if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, buffer, sizeof(buffer), 1000)) < 0)
        return rval;
    if(buffer[1] != USB_DT_STRING)
        return 0;
    if((unsigned char)buffer[0] < rval)
        rval = (unsigned char)buffer[0];
    rval /= 2;
    /* lossy conversion to ISO Latin1 */
    for(i=1;i<rval;i++){
        if(i > buflen)  /* destination buffer overflow */
            break;
        buf[i-1] = buffer[2 * i];
        if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
            buf[i-1] = '?';
    }
    buf[i-1] = 0;
    return i-1;
}


/* PowerSwitch uses the free shared default VID/PID. If you want to see an
 * example device lookup where an individually reserved PID is used, see our
 * RemoteSensor reference implementation.
 */

#define USB_ERROR_NOTFOUND  1
#define USB_ERROR_ACCESS    2
#define USB_ERROR_IO        3

static int usbOpenDevice(usb_dev_handle **device, int vendor, char *vendorName, int product, char *productName)
{
struct usb_bus      *bus;
struct usb_device   *dev;
usb_dev_handle      *handle = NULL;
int                 errorCode = USB_ERROR_NOTFOUND;
static int          didUsbInit = 0;

    if(!didUsbInit){
        didUsbInit = 1;
        usb_init();
    }
    usb_find_busses();
    usb_find_devices();
    for(bus=usb_get_busses(); bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next){
            if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product){
                char    string[256];
                int     len;
                handle = usb_open(dev); /* we need to open the device in order to query strings */
                if(!handle){
                    errorCode = USB_ERROR_ACCESS;
                    fprintf(stderr, "Warning: cannot open USB device: %s\n", usb_strerror());
                    continue;
                }
                if(vendorName == NULL && productName == NULL){  /* name does not matter */
                    break;
                }
                /* now check whether the names match: */
                len = usbGetStringAscii(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
                if(len < 0){
                    errorCode = USB_ERROR_IO;
                    fprintf(stderr, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
                }else{
                    errorCode = USB_ERROR_NOTFOUND;
                    /* fprintf(stderr, "seen device from vendor ->%s<-\n", string); */
                    if(strcmp(string, vendorName) == 0){
                        len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
                        if(len < 0){
                            errorCode = USB_ERROR_IO;
                            fprintf(stderr, "Warning: cannot query product for device: %s\n", usb_strerror());
                        }else{
                            errorCode = USB_ERROR_NOTFOUND;
                            /* fprintf(stderr, "seen product ->%s<-\n", string); */
                            if(strcmp(string, productName) == 0)
                                break;
                        }
                    }
                }
                usb_close(handle);
                handle = NULL;
            }
        }
        if(handle)
            break;
    }
    if(handle != NULL){
        errorCode = 0;
        *device = handle;
    }
    return errorCode;
}


int main(int argc, char **argv)
{
usb_dev_handle      *handle = NULL;
unsigned char       buffer[256];
int                 nBytes;

    if(argc < 2){
        usage(argv[0]);
        exit(1);
    }
    usb_init();
    if(usbOpenDevice(&handle, USBDEV_SHARED_VENDOR, "www.tahina.priv.at", USBDEV_SHARED_PRODUCT, "Sensor") != 0){
        fprintf(stderr, "Could not find USB device \"Sensor\" with vid=0x%x pid=0x%x\n", USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
        exit(1);
    }
/* We have searched all devices on all busses for our USB device above. Now
 * try to open it and perform the vendor specific control operations for the
 * function requested by the user.
 */
    if(strcmp(argv[1], "test") == 0){
        int i, v, r;
/* The test consists of writing 1000 random numbers to the device and checking
 * the echo. This should discover systematic bit errors (e.g. in bit stuffing).
 */
        for(i=0;i<1000;i++){
            v = rand() & 0xffff;
            nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_ECHO, v, 0, (char *)buffer, sizeof(buffer), 5000);
            if(nBytes < 2){
                if(nBytes < 0)
                    fprintf(stderr, "USB error: %s\n", usb_strerror());
                fprintf(stderr, "only %d bytes received in iteration %d\n", nBytes, i);
                exit(1);
            }
            r = buffer[0] | (buffer[1] << 8);
            if(r != v){
                fprintf(stderr, "data error: received 0x%x instead of 0x%x in iteration %d\n", r, v, i);
                exit(1);
            }
        }
        printf("test succeeded\n");
    }else if(strcmp(argv[1], "status") == 0){
        int i,j;
        nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_STATUS, 0, 0, (char *)buffer, sizeof(buffer), 5000);
	printf("got %d bytes of status:\n", nBytes);
	for(i = 0; i<nBytes; i++)
	  printf("%02d: %3d (0x%02x)\n", i, buffer[i], buffer[i]);
	j = (buffer[0] << 8) | buffer[1];
	for(i = 0; i<12; i++) {
	  if(j & 0x8000)
	    printf("%s ", statusbits[i]);
	  j<<=1;
	}
	printf("\nOffset = %d\n", buffer[1] & 0xf);
    }else if(strcmp(argv[1], "data") == 0) {
      int i;
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_DATA, 0, 0, (char *)buffer, sizeof(buffer), 5000);
      printf("got %d bytes of data:\n", nBytes);
      printf("state: %s\n", states[buffer[0]]);
      printf("len:   %d\n", buffer[1]);
      printf("cksum: %d\n", buffer[2]);
      printf("i:     %d\n", buffer[3]);
      printf("timeo: %d\n", (buffer[4]<<8)|buffer[5]);
      printf("\n");
      #if 0
      for(i = 0; i < nBytes; i++)
	printf("%02d %3d\n", i, buffer[i]);
      printf("\n");
      #endif
      printf("type:  %d\n", ((struct data *)(buffer + 6))->type);
      printf("sender:%d\n", ((struct data *)(buffer + 6))->sender);
      printf("seq:   %d\n", ((struct data *)(buffer + 6))->seq);
      printf("volt:  %.2f\n", 1.30/(((struct data *)(buffer + 6))->voltage/1024.0));
    } else if(strcmp(argv[1], "stats") == 0) {
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_STATS, 0, 0, (char *)buffer, sizeof(buffer), 5000);
      printf("stats:\n");
      printf("OK:              %d\n", ((struct stats *)buffer)->ok);
      printf("cksum:           %d\n", ((struct stats *)buffer)->e_cksum);
      printf("TO reading:      %d\n", ((struct stats *)buffer)->e_to_reading);
      printf("TO reading done: %d\n",
	     ((struct stats *)buffer)->e_to_reading_done);
      printf("bad len:         %d\n", ((struct stats *)buffer)->e_bad_len);
    }
    usb_close(handle);
    return 0;
}
